#include "PhysicalObjects/chirp.h"

#include <Box2D/Box2D.h>

#include <QPainter>
#include <QDebug>
#include <QGraphicsScene>
#include <QSharedPointer>
#include <QJsonObject>
#include <QJsonArray>

#include <limits>
#include <cmath>

#include "physicalobjectfactory.h"
#include "mathutility.h"
#include "drawutilities.h"
#include "irsensor.h"
#include "world.h"
#include "robotrepository.h"
#include "chirpnetworkcontroller.h"
#include "jsonhelpers.h"

PhysicalObject *chirpFactory(World* world, const QJsonObject &obj)
{
    quint32 id = obj["id"].toDouble();
    float angle = 0.0f;
    if(obj.contains("angle"))
        angle = obj["angle"].toDouble();

    QPointF position = QPointF(0,0);
    if(obj.contains("position"))
        position = jsonObjectToPoint(obj["position"].toObject());

    float noiseMax = 0.1f;
    if(obj.contains("noisemax"))
        noiseMax = obj["noisemax"].toDouble();

    Chirp *dw = new Chirp(world->world(), id, position, angle, noiseMax, 5.0f, 5.0f, 2.0f, 0.5f);
    RobotRepository::instance()->addRobot(dw);

    return dw;
}

Chirp::Chirp(b2World *world, unsigned id, QPointF position, float startAngle, float noiseMax, float radius, float wheelDistance, float wheelRadius, float wheelWidth) : QGraphicsItem()
{
    Robot::setId(id);
    Robot::setController(new ChirpNetworkController(const_cast<Chirp*>(this)));

    _radius = radius;
    _wheelRadius = wheelRadius;
    _wheelWidth = wheelWidth;
    _halfWheelDistance = wheelDistance / 2.0f;
    _noiseMax = noiseMax;

    createPhysicalBody(position, startAngle, world);

    computeBoundingRect();

    createSensors();

    _leftWheelSpeed = 0.0f;
    _rightWheelSpeed = 0.0f;

    setPos(position);
    setRotation(startAngle);

    setFlag(QGraphicsItem::ItemIsSelectable);
}

Chirp::~Chirp()
{
    for(int i = 0; i < _sensors.size(); ++i)
        delete _sensors[i];

    _body->GetWorld()->DestroyBody(_body);
}

QRectF Chirp::boundingRect() const
{
    return _boundingRect;
}

void Chirp::paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *widget)
{
    // Removes two annoying warnings during build.
    //Q_UNUSED(option);
    Q_UNUSED(widget);
    painter->save();
    painter->setClipRect(option->exposedRect);
    QPen p(QBrush(Qt::black), 0.1f);
    painter->setPen(p);

    painter->drawLine(QPointF(0.0f,_radius), QPointF(0.0f, _radius / 2.0f));

    // Left wheel.
    painter->drawEllipse(QPointF(0,0), _radius, _radius);
    QRectF r1(QPointF(-_halfWheelDistance - _wheelWidth, _wheelRadius), QPointF(-_halfWheelDistance + _wheelWidth, -_wheelRadius));
    painter->drawRect(r1);

    // Right wheel.
    QRectF r2(QPointF(_halfWheelDistance - _wheelWidth, _wheelRadius), QPointF(_halfWheelDistance + _wheelWidth, -_wheelRadius));
    painter->drawRect(r2);
    if(isSelected())
    {
        QPen p2(Qt::red);
        p2.setWidthF(0.1f);
        painter->setPen(p2);
        painter->drawRect(boundingRect());
    }
    painter->restore();
}

void Chirp::step(float dt)
{
    Q_UNUSED(dt);

    applyImpulses();
    removeLateralSpeed();

    foreach (Sensor *sensor, _sensors)
    {
        sensor->update();
    }
}

void Chirp::stepGraphics()
{
    b2Vec2 pos = _body->GetPosition();
    float angle = _body->GetAngle() * RADTODEG;

    setPos(pos.x, pos.y);
    setRotation(angle);
}

/*
 * applyImpulses calculates the difference between desired speed and actual speed, then applies
 * an impulse to make the robot move at the desired speed.
 * Function is called every frame through advance().
 *
 * See http://www.iforce2d.net/b2dtut/top-down-car for an in-depth explanation of the code.
 */
void Chirp::applyImpulses()
{
    b2Vec2 forceDirection = _body->GetWorldVector(b2Vec2(0, 1));

    float leftWheelDot =  b2Dot(forceDirection, _body->GetLinearVelocityFromLocalPoint(b2Vec2(-_halfWheelDistance, 0)));
    float rightWheelDot = b2Dot(forceDirection, _body->GetLinearVelocityFromLocalPoint(b2Vec2(_halfWheelDistance, 0)));

    // Calculate amount of speed in the wheel direction.
    float leftNoise = _leftWheelSpeed * (1 - ((float)rand() / RAND_MAX)*2) * _noiseMax;
    float rightNoise = _rightWheelSpeed * (1 - ((float)rand() / RAND_MAX)*2) * _noiseMax;
    b2Vec2 leftSpeed = leftWheelDot * forceDirection;
    b2Vec2 rightSpeed = rightWheelDot * forceDirection;

    // Difference between current speed in wheel direction and desired speed.
    b2Vec2 leftDelta = ((_leftWheelSpeed + leftNoise) * forceDirection) - leftSpeed;
    b2Vec2 rightDelta = ((_rightWheelSpeed + rightNoise) * forceDirection) - rightSpeed;

    // Necessary because of numerical instability with floating point numbers.
    if(fabs(leftDelta.Length()) > std::numeric_limits<float>::epsilon())
    {
        b2Vec2 leftWheelPosition = _body->GetWorldPoint(b2Vec2(-_halfWheelDistance, 0));
        b2Vec2 leftWheelImpuse = _body->GetMass() * 0.5f * leftDelta;
        _body->ApplyLinearImpulse(leftWheelImpuse, leftWheelPosition);
    }
    if(fabs(rightDelta.Length()) > std::numeric_limits<float>::epsilon())
    {
        b2Vec2 rightWheelPosition = _body->GetWorldPoint(b2Vec2(_halfWheelDistance, 0));
        b2Vec2 rightWheelImpuse = _body->GetMass() * 0.5f * rightDelta;
        _body->ApplyLinearImpulse(rightWheelImpuse, rightWheelPosition);
    }
}

/*
 * removeLateralSpeed finds the lateral speed vector of the robot and applies the inverse vector as an instant impulse
 * to center of the body. This makes the robot appear to have proper friction at the wheels, instead of sliding around like it was on a slippery surface.
 * Function is called every frame through advance().
 *
 * Note: Due to how collisions work, robots are able to slightly nudge each other sideways through separation after collision. The implementation here has
 *       intentionally left out the skidding part of the original source.
 *
 * See http://www.iforce2d.net/b2dtut/top-down-car for an in-depth explanation of the code.
 */
void Chirp::removeLateralSpeed()
{
    b2Vec2 currentRightNormal = _body->GetWorldVector(b2Vec2(1, 0));
    b2Vec2 lateralVelocity = b2Dot(currentRightNormal, _body->GetLinearVelocity()) * currentRightNormal;
    b2Vec2 impulse = _body->GetMass() * -lateralVelocity;
    _body->ApplyLinearImpulse(impulse, _body->GetWorldCenter());
}

/*
 * computeBoundingRect is a helper function for calculating the bounding rectangle of the robot on screen.
 *
 * Called by constructor as a helper function.
 */
void Chirp::computeBoundingRect()
{
    long maxValue = std::numeric_limits<int>::max();
    long minValue = std::numeric_limits<int>::min();
    _boundingRect = QRectF(maxValue, maxValue, minValue, minValue);

    // Go through each shape and calculate the bounding box, stretching _boundingRect to encompass all the shapes of the body.
    b2Fixture* fixture = _body->GetFixtureList();
    while(fixture != NULL)
    {
        b2AABB aabb = fixture->GetAABB(0);
        b2Vec2 extents = aabb.GetExtents();
        b2Vec2 center = aabb.GetCenter() - _body->GetPosition();
        _boundingRect.setLeft(std::min<float>(_boundingRect.left(), -extents.x - center.x));
        _boundingRect.setTop(std::min<float>(_boundingRect.top(), -extents.y - center.y));

        _boundingRect.setRight(std::max<float>(_boundingRect.right(), extents.x - center.x));
        _boundingRect.setBottom(std::max<float>(_boundingRect.bottom(), extents.y - center.y));

        fixture = fixture->GetNext();
    }
}

/*
 * createPhysicalBody is a helper function used by the constructor for creating the physical body of the object,
 * putting it into the simulated world sent in as a parameter.
 */
void Chirp::createPhysicalBody(QPointF position, float startAngle, b2World* world)
{
    // Create physical body for simulation. This contains no actual shapes, but is just a "container" for holding fixtures (shapes).
    b2BodyDef bodyDef;
    bodyDef.type = b2_dynamicBody;
    bodyDef.position.Set(position.x(), position.y());
    bodyDef.angle = startAngle;
    _body = world->CreateBody(&bodyDef);
    _body->SetSleepingAllowed(true);

    // Create main body shape (circle) and attach it to simulation body.
    b2CircleShape bodyShape;
    bodyShape.m_p.Set(0, 0);
    bodyShape.m_radius = _radius;
    b2FixtureDef bodyFixtureDef;
    bodyFixtureDef.shape = &bodyShape;
    bodyFixtureDef.density = 10;
    _body->CreateFixture(&bodyFixtureDef);

    // Create left wheel shape and attach it to body with no mass.
    b2PolygonShape leftWheelShape;
    leftWheelShape.SetAsBox(_wheelWidth, _wheelRadius * 2, b2Vec2(-_halfWheelDistance, 0.0f), 0);
    b2FixtureDef leftWheelFixtureDef;
    leftWheelFixtureDef.shape = &leftWheelShape;
    leftWheelFixtureDef.density = 0;
    _body->CreateFixture(&leftWheelFixtureDef);

    // Create right wheel shape and attach it to body with no mass.
    b2PolygonShape rightWheelShape;
    rightWheelShape.SetAsBox(_wheelWidth, _wheelRadius * 2, b2Vec2(_halfWheelDistance, 0.0f), 0);
    b2FixtureDef rightWheelFixtureDef;
    rightWheelFixtureDef.shape = &rightWheelShape;
    rightWheelFixtureDef.density = 0;
    _body->CreateFixture(&rightWheelFixtureDef);

    _body->SetAngularDamping(0.0f);
    _body->SetLinearDamping(0.0f);

    _body->SetUserData(this);
}

void Chirp::createSensors()
{
    const float angleStep = (360.0f / 8.0f);
    _sensors.resize(8);
    // Create sensors evenly spaced around the robot.
    for(int i = 0; i < 8; i++)
    {
        float angle = angleStep * i;
        QPointF position = QPointF(sinf(angle * DEGTORAD) * (_radius - 0.1f), cosf(angle * DEGTORAD) * (_radius - 0.1f));
        ChirpIrSensor* sensor = new ChirpIrSensor(this, position, angle, 25, 1023, 40, 1);
        _sensors[i] = sensor;
    }
}


void Chirp::setSpeed(QVector<float> speeds)
{
    //QWriteLocker locker(&_readWriteLock);
    _leftWheelSpeed = speeds[0];
    _rightWheelSpeed = speeds[1];
}


QVector<float> Chirp::sensors() const
{
    QVector<float> sensorValues = QVector<float>(_sensors.size());
    for(int i = 0; i < _sensors.size(); ++i)
    {
        sensorValues[i] = _sensors[i]->value();
    }
    return sensorValues;
}
