#include "PhysicalObjects/pushableobject.h"
#include "mathutility.h"
#include "drawutilities.h"
#include "jsonhelpers.h"

#include <QDebug>
#include <QJsonArray>

PhysicalObject *pushableObjectFactory(World *world, const QJsonObject &obj)
{
    QList< QList< QPointF > > hulls;
    QJsonArray jsonVertices = obj["hulls"].toArray();
    foreach (const QJsonValue &points, jsonVertices)
    {
        QList< QPointF > vertices;
        foreach(const QJsonValue &point, points.toArray())
        {
            vertices << jsonObjectToPoint(point.toObject());
        }
        hulls << vertices;
    }

    float density = 1.0f;
    if(obj["density"].isDouble())
        density = obj["density"].toDouble();

    QPointF position = QPointF(0,0);
    if(obj["position"].isObject())
        position = jsonObjectToPoint(obj["position"].toObject());

    float angle = 0.0f;
    if(obj["angle"].isDouble())
        angle = obj["angle"].toDouble();

    return new PushableObject(world->world(), hulls, density, position, angle);
}

PushableObject::PushableObject(b2World *world, QList<QList<QPointF> > hulls, float density, QPointF position, float angle) :
    PhysicalObject(), QGraphicsItem()
{
    b2BodyDef bodyDef;
    bodyDef.position.Set(position.x(), position.y());
    bodyDef.angle = angle * DEGTORAD;
    bodyDef.type = b2_dynamicBody;

    _body = world->CreateBody(&bodyDef);
    _body->SetLinearDamping(0.3f);
    _body->SetAngularDamping(0.3f);
    foreach(const QList< QPointF > &vertices, hulls)
    {
        b2PolygonShape *shape = new b2PolygonShape();
        b2Vec2 *points = new b2Vec2[vertices.length()];
        QPointFTob2Vec2(vertices, points);
        shape->Set(points, vertices.length());
        b2Fixture *fixture = _body->CreateFixture(shape, density);
        fixture->SetFriction(0.3f);
    }

    ComputeBoundingRect();

    setPos(position);
    setRotation(angle);
}

PushableObject::~PushableObject()
{
}

QRectF PushableObject::boundingRect() const
{
    return _boundingRect;
}

void PushableObject::paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *widget)
{
    genericPaint(painter, option, widget, _body);
}

void PushableObject::step(float dt)
{
    Q_UNUSED(dt);

    // TODO: Implement applyFriction
    applyFriction();
}

void PushableObject::stepGraphics()
{
    b2Vec2 pos = _body->GetPosition();
    float angle = _body->GetAngle() * RADTODEG;
    setPos(pos.x, pos.y);
    setRotation(angle);
}

void PushableObject::ComputeBoundingRect()
{
    long maxValue = 50000;
    long minValue = -50000;
    _boundingRect = QRectF(maxValue, maxValue, minValue, minValue);
    b2CircleShape c;
    b2PolygonShape p;

    b2Fixture* fixture = _body->GetFixtureList();
    while(fixture != NULL)
    {
        b2AABB aabb = fixture->GetAABB(0);
        b2Vec2 extents = aabb.GetExtents();
        b2Vec2 center = aabb.GetCenter() - _body->GetPosition();
        _boundingRect.setLeft(std::min<float>(_boundingRect.left(), -extents.x - center.x));
        _boundingRect.setTop(std::min<float>(_boundingRect.top(), -extents.y - center.y));

        _boundingRect.setRight(std::max<float>(_boundingRect.right(), extents.x - center.x));
        _boundingRect.setBottom(std::max<float>(_boundingRect.bottom(), extents.y - center.y));

        fixture = fixture->GetNext();
    }
}

void PushableObject::applyFriction()
{
    // TODO: Implement static friction and dynamic friction.
}
