#include "PhysicalObjects/wall.h"
#include "mathutility.h"
#include "drawutilities.h"
#include "jsonhelpers.h"

#include <QDebug>
#include <QJsonArray>

PhysicalObject *wallFactory(World *world, const QJsonObject &obj)
{
    QList< QList< QPointF > > hulls;
    QJsonArray jsonVertices = obj["hulls"].toArray();
    foreach (const QJsonValue &points, jsonVertices)
    {
        QList< QPointF > vertices;
        foreach(const QJsonValue &point, points.toArray())
        {
            vertices << jsonObjectToPoint(point.toObject());
        }
        hulls << vertices;
    }

    QPointF position = QPointF(0,0);
    if(obj["position"].isObject())
        position = jsonObjectToPoint(obj["position"].toObject());

    float angle = 0.0f;
    if(obj["angle"].isDouble())
        angle = obj["angle"].toDouble();

    return new Wall(world->world(), hulls, position, angle);
}

Wall::Wall(b2World* world, QList< QList< QPointF > > hulls, QPointF position, float angle) :
    PhysicalObject(), QGraphicsItem()
{
    b2BodyDef bodyDef;
    bodyDef.position.Set(position.x(), position.y());
    bodyDef.angle = angle * DEGTORAD;
    bodyDef.type = b2_staticBody;

    _body = world->CreateBody(&bodyDef);

    foreach(const QList< QPointF > &vertices, hulls)
    {
        b2PolygonShape *shape = new b2PolygonShape();
        b2Vec2 *points = new b2Vec2[vertices.length()];
        QPointFTob2Vec2(vertices, points);
        shape->Set(points, vertices.length());
        b2Fixture *fixture = _body->CreateFixture(shape, 0);
        fixture->SetFriction(0.3f);
    }

    ComputeBoundingRect();

    setPos(position);
    setRotation(angle);
}

Wall::~Wall()
{
}

QRectF Wall::boundingRect() const
{
    return _boundingRect;
}

void Wall::paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *widget)
{
    genericPaint(painter, option, widget, _body);
}

void Wall::ComputeBoundingRect()
{
    long maxValue = 50000;
    long minValue = -50000;
    _boundingRect = QRectF(maxValue, maxValue, minValue, minValue);
    b2CircleShape c;
    b2PolygonShape p;

    b2Fixture* fixture = _body->GetFixtureList();
    while(fixture != NULL)
    {
        b2AABB aabb = fixture->GetAABB(0);
        b2Vec2 extents = aabb.GetExtents();
        b2Vec2 center = aabb.GetCenter() - _body->GetPosition();
        _boundingRect.setLeft(std::min<float>(_boundingRect.left(), -extents.x - center.x));
        _boundingRect.setTop(std::min<float>(_boundingRect.top(), -extents.y - center.y));

        _boundingRect.setRight(std::max<float>(_boundingRect.right(), extents.x - center.x));
        _boundingRect.setBottom(std::max<float>(_boundingRect.bottom(), extents.y - center.y));

        fixture = fixture->GetNext();
    }
}
