#ifndef ARENA_H
#define ARENA_H

#include <QString>
#include <QJsonObject>
#include <QGraphicsItem>
#include <Box2D/Box2D.h>

#include "world.h"
#include "physicalobject.h"

class Arena;

/*!
 * \brief chirpFactory follows the PhysicalObject FactoryFunction, creating the Arena object from a QJsonObject. \see physicalobjectfactory.h
 * \param world World to put the physical object model in.
 * \param obj JsonObject containing the parameters for the object.
 * \return A constructed Arena object.
 */
PhysicalObject *arenaFactory(World *world, const QJsonObject &obj);

/*!
 * \brief The Arena class creates four walls that are immovable at a specified distance from the center position.
 *
 * The Arena class creates four walls that are immovable at a specified distance from the center position. It handles creation of all four walls in the Box2D world, along with drawing them
 * in the simulator. It's a handy class for restricting movement in the world.
 */
class Arena : public PhysicalObject, public QGraphicsItem
{
public:
    /*!
     * \brief Constructor creating an Arena object based on a number of parameters.
     * \param world World for the arena to be created in.
     * \param position Center position of the arena.
     * \param halfWidth Distance from center position to both walls on the x-axis.
     * \param halfHeight Distance from the center position to both walls on the y-axis.
     *
     * Creates a physical model of the arena with Box2D and sets up the drawing of the walls.
     */
    explicit Arena(b2World *world, QPointF position = QPointF(0,0), float halfWidth = 200, float halfHeight = 200);
    ~Arena();

    /*!
     * \brief boundingRect returns the computed bounding rectangle used for drawing and redrawing the object. From the QGraphicsItem abstract base class.
     * \return bounding rectangle.
     */
    QRectF boundingRect() const;

    /*!
     * \brief paint Draws the walls for display. From the QGraphicsItem abstract base class.
     */
    void paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *widget);

    /*!
     * \brief Simple getter for _body field. From the PhysicalObject interface class.
     * \return Physical representation of object used by b2World class.
     */
    b2Body *body() const { return _body; }

    /*!
     * \brief Simple getter for returning QGraphicsItem connected to this object. From the PhysicalObject interface class.
     * \return this casted to a non-const pointer.
     */
    QGraphicsItem *item() const { return const_cast<Arena*>(this); }

signals:
    
public slots:
    
private:
    void CreatePhysicalBody(QPointF position, float width, float height, b2World *world);

private:
    friend PhysicalObject *arenaFactory(World *world, const QJsonObject &obj);
    QRectF _boundingRect;

    b2Body* _body;

    static const QString Name;
};

#endif // ARENA_H
