#ifndef CHIRP_H
#define CHIRP_H

#include <QGraphicsItem>
#include <QSharedPointer>
#include <QVector>

#include <QPainter>
#include <QStyleOptionGraphicsItem>
#include <QWidget>
#include <QGraphicsScene>
#include <QJsonObject>
#include <QReadWriteLock>

#include "robot.h"
#include "irsensor.h"
#include "world.h"
#include "physicalobject.h"

class Chirp;

/*!
 * \brief chirpFactory follows the PhysicalObject FactoryFunction, creating the Chirp object from a JSON object. \see physicalobjectfactory.h
 * \param world World to put the physical object model in.
 * \param obj JSON object containing the parameters for the robot.
 * \return A constructed Chirp object.
 */
PhysicalObject *chirpFactory(World *world, const QJsonObject &obj);

/*!
 * \brief The Chirp class implements the simulation model of the Chirp robot.
 *
 * It implements the Robot interface, the PhysicalObject abstract base class and
 * the QGraphicsItem abstract base class. It handles the custom physics required to make the robot not slide like the wheels have no sideways friction,
 * drawing the main body and both wheels, and lastly holds a ChirpController that can control the speeds of the wheels along with reading sensor datas.
 *
 * The Chirp robot in simulation:
 * \image html chirp.png
 */
class Chirp : public Robot, public PhysicalObject, public QGraphicsItem
{
public:
    /*!
     * \brief Constructor creating a Chirp object based on a number of parameters.
     * \param world World for the robot to be created in.
     * \param id Unique id for the robot.
     * \param position Starting position of the robot.
     * \param startAngle Starting angle of the robot.
     * \param noiseMax Max noise values on the wheel speeds.
     * \param radius Radius of the robot body.
     * \param wheelDistance Distance between wheels.
     * \param wheelRadius Radius of the wheels.
     * \param wheelWidth Width of the wheels.
     *
     * Creates a physical model of the robot with Box2D, sets up the drawing of the robot and sets up the sensors of the robot.
     */
    explicit Chirp::Chirp(b2World *world, unsigned id, QPointF position, float startAngle, float noiseMax, float radius, float wheelDistance, float wheelRadius, float wheelWidth);
    ~Chirp();

    /*!
     * \brief Returns the computed bounding rectangle used for drawing and redrawing the object.
     * \return bounding rectangle.
     */
    QRectF boundingRect() const;

    /*!
     * \brief Draws the robot and wheels for display.
     */
    void paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *widget);

    /*!
     * \brief Simple setter for wheel speeds.
     * \param speeds Contains two floats, first for left wheel and second for right wheel.
     */
    void setSpeed(QVector<float> speeds);

    /*!
     * \brief Getter for all the sensors on the robot.
     * \return Array containing sensor values read at the time of calling.
     */
    QVector<float> sensors() const;

    /*!
     * \brief Simple getter for _body field.
     * \return Physical representation of object used by b2World class.
     */
    b2Body *body() const { return _body; }

    /*!
     * \brief Simple getter for returning QGraphicsItem connected to this object.
     * \return this casted to a non-const pointer.
     */
    QGraphicsItem *item() const { return const_cast<Chirp*>(this); }

    /*!
     * \brief This is called by World each step to help implement more custom behaviors in the simulation (like wheels not sliding sideways) and
     *        to update the sensor values.
     * \param dt The time in seconds since last step.
     *
     *
     */
    void step(float dt);

    /*!
     * \brief Updates the viewable model of the robot to reflect the current world state.
     */
    void stepGraphics();

private:
    friend PhysicalObject *chirpFactory(World *world, const QJsonObject &obj);

    // Helper functions for advance()
    void applyImpulses();
    void removeLateralSpeed();

    // Helper functions for constructor
    void computeBoundingRect();
    void createPhysicalBody(QPointF position, float startAngle, b2World* world);
    void createSensors();
private:
    b2Body* _body;
    float _halfWheelDistance;
    float _wheelRadius;
    float _wheelWidth;
    float _radius;
    float _noiseMax;

    quint32 _id;

    QRectF _boundingRect;

    QVector<Sensor*> _sensors;

    float _leftWheelSpeed;
    float _rightWheelSpeed;

    bool _isSelected;
};

#endif // CHIRP_H
