#ifndef PUSHABLEOBJECT_H
#define PUSHABLEOBJECT_H

#include <QJsonObject>
#include <QGraphicsItem>
#include <QRectF>
#include <QPainter>
#include <Box2D/Box2D.h>

#include "physicalobject.h"
#include "world.h"

class PushableObject;

/*!
 * \brief chirpFactory follows the PhysicalObject FactoryFunction, creating the PushableObject object from a QJsonObject. \see physicalobjectfactory.h
 * \param world World to put the physical object model in.
 * \param obj JsonObject containing the parameters for the object.
 * \return A constructed PushableObject object.
 */
PhysicalObject *pushableObjectFactory(World *world, const QJsonObject &obj);

/*!
 * \brief The PushableObject class represents a physical object that other objects can push around.
 *
 * The PushableObject class represents a physical object that other objects can push around. It implements a very basic friction model to
 * stop it from just sliding off if nudged at all.
 *
 * The pushable object in simulation:
 * \image html wall.png
 */
class PushableObject : public PhysicalObject, public QGraphicsItem
{
public:
    /*!
     * \brief Constructor creating an PushableObject object based on a number of parameters.
     * \param world World for the pushable object to be created in.
     * \param hulls List of hulls forming the shape of the object. Each hull forms a separate collision mesh, allowing simulation of a concave hull through
     *              making them a series of convex hulls. \note Vertices must be in counterclockwise order, and must form a convex hull.
     * \param density Density of the object. Decides the mass of the body based on the area of the shapes.
     * \param position Starting position of the pushable object.
     * \param angle Starting angle of the pushable object.
     *
     * Creates a physical model of the pushable object with Box2D and sets up the drawing of the object.
     */
    explicit PushableObject(b2World *world, QList< QList< QPointF > > hulls, float density, QPointF position, float angle);
    ~PushableObject();

    /*!
     * \brief boundingRect returns the computed bounding rectangle used for drawing and redrawing the object. From the QGraphicsItem abstract base class.
     * \return bounding rectangle.
     */
    QRectF boundingRect() const;

    /*!
     * \brief paint Draws the object for display. From the QGraphicsItem abstract base class.
     */
    void paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *widget);

    /*!
     * \brief Simple getter for _body field. From the PhysicalObject interface class.
     * \return Physical representation of object used by b2World class.
     */
    b2Body *body() const { return _body; }

    /*!
     * \brief Simple getter for returning QGraphicsItem connected to this object. From the PhysicalObject interface class.
     * \return this casted to a non-const pointer.
     */
    QGraphicsItem *item() const { return const_cast<PushableObject*>(this); }

    /*!
     * \brief step updates the position of the object as it gets pushed around.
     * \param dt time in seconds passed since last step.
     */
    void step(float dt);

    /*!
     * \brief Updates the viewable model of the object to reflect the current world state.
     */
    void stepGraphics();
private:
    friend PhysicalObject *pushableObjectFactory(World *world, const QJsonObject &obj);
    void ComputeBoundingRect();

    void applyFriction();
private:
    b2Body* _body;

    QRectF _boundingRect;
    
};

#endif // PUSHABLEOBJECT_H
