#ifndef WALL_H
#define WALL_H

#include <QJsonObject>
#include <QGraphicsItem>
#include <QRectF>
#include <QPainter>
#include <Box2D/Box2D.h>

#include "physicalobject.h"
#include "world.h"

class Wall;

/*!
 * \brief chirpFactory follows the PhysicalObject FactoryFunction, creating the Wall object from a QJsonObject. \see physicalobjectfactory.h
 * \param world World to put the physical object model in.
 * \param obj JsonObject containing the parameters for the object.
 * \return A constructed Wall object.
 */
PhysicalObject *wallFactory(World *world, const QJsonObject &obj);

/*!
 * \brief The Wall class represents a physical object that is immovable during simulation.
 *
 * The Wall class represents a physical object that is immovable during simulation. It is perfect for creating an obstacle that cannot be moved.
 *
 * The immovable wall object in simulation:
 * \image html wall.png
 */
class Wall : public PhysicalObject, public QGraphicsItem
{
public:
    /*!
     * \brief Constructor creating a Wall object based on a number of parameters.
     * \param world World for the pushable object to be created in.
     * \param hulls List of hulls forming the shape of the object. Each hull forms a separate collision mesh, allowing simulation of a concave hull through
     *              making them a series of convex hulls. \note Vertices must be in counterclockwise order, and must form a convex hull.
     * \param position Starting position of the pushable object.
     * \param angle Starting angle of the pushable object.
     *
     * Creates a physical model of the wall object with Box2D and sets up the drawing of the object.
     */
    explicit Wall(b2World *world, QList< QList< QPointF > > hulls, QPointF position = QPointF(0,0), float angle = 0);
    ~Wall();

    /*!
     * \brief boundingRect returns the computed bounding rectangle used for drawing and redrawing the object. From the QGraphicsItem abstract base class.
     * \return bounding rectangle.
     */
    QRectF boundingRect() const;

    /*!
     * \brief paint Draws the wall for display. From the QGraphicsItem abstract base class.
     */
    void paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *widget);

    /*!
     * \brief Simple getter for _body field. From the PhysicalObject interface class.
     * \return Physical representation of object used by b2World class.
     */
    b2Body *body() const { return _body; }

    /*!
     * \brief Simple getter for returning QGraphicsItem connected to this object. From the PhysicalObject interface class.
     * \return this casted to a non-const pointer.
     */
    QGraphicsItem *item() const { return const_cast<Wall*>(this); }
private:
    friend PhysicalObject *wallFactory(World *world, const QJsonObject &obj);
    void ComputeBoundingRect();

private:
    b2Body* _body;

    QRectF _boundingRect;
    
};

#endif // WALL_H
