#ifndef CHIRPNETWORKCONTROLLER_H
#define CHIRPNETWORKCONTROLLER_H

#include <QVector>
#include <QTextStream>

#include "PhysicalObjects/chirp.h"
#include "controller.h"

class Chirp;

/*!
 * \brief The ChirpNetworkController class handles the controller messages the Chirp robot accepts.
 *
 * The ChirpNetworkController class handles the controller messages the Chirp robot accepts. The two current messages allowed are
 * the "setspeed"-message and the "getsensors"-message. The former lets controllers set the speed of each wheel on the robot, effectively
 * controlling the robot's movement. The latter lets controllers read the distance sensors on the robot.
 */
class ChirpNetworkController : public Controller
{
    Q_OBJECT
public:
    /*!
     * \brief Constructs a ChirpNetworkController object and sets its assigned robot.
     * \param robot Robot this controller is assigned to.
     */
    ChirpNetworkController(Chirp *robot);

    /*!
     * \brief Main entry point for the controller. Handles setspeed and getsensor messages.
     * \param stream QTextStream wrapping the message sent over UDP. Allows for partial consumation of message.
     * \return Utf8 string with either the expected response from the message handler, or a utf8 string containing "error"
     *         to mark that a failure occured.
     *
     * The full setspeed message as sent over UDP follows this template:
     *      <blockquote>&lt;id&gt; setspeed &lt;left wheel speed&gt; &lt;right wheel speed&gt;</blockquote>
     * return
     *      <blockquote>ok</blockquote>
     * <br />
     *
     * The full getsensors message as sent over UDP follows this template:
     *      <blockquote>&lt;id&gt; getsensors</blockquote>
     * return
     *      <blockquote>ok &lt;sensor 1&gt; &lt;sensor 2&gt; ... &lt;sensor 8&gt;</blockquote>
     */
    virtual QByteArray handleMessage(QTextStream &stream);

private:
    QByteArray handleSetSpeedMessage(QTextStream &stream);
    QByteArray handleGetSensorsMessage();

private:
    quint32 _id;
    Chirp *_robot;
    QVector<QString> _acceptedMessages;
};

#endif // NETWORKCONTROLLER_H
