#ifndef IRSENSOR_H
#define IRSENSOR_H

#include <QGraphicsItem>
#include <QVector>
#include <QJsonObject>
#include <QPointF>
#include <Box2D/Box2D.h>

#include "sensor.h"
#include "world.h"
#include "robot.h"
#include "PhysicalObjects/physicalobject.h"


/*!
 * \brief The IrSensorRayCastCallback class implements the callback class used by Box2D to report results of raycasting.
 *
 * This enables the use of the more efficient b2World.RayCast method over the less efficient b2Fixture.RayCast method.
 */
class IrSensorRayCastCallback : public b2RayCastCallback
{
public:
    IrSensorRayCastCallback() : b2RayCastCallback(), _fraction(1) {}
    float32 ReportFixture(b2Fixture *fixture, const b2Vec2 &point, const b2Vec2 &normal, float32 fraction);

    /*!
     * \brief reset the fraction and object hit.
     */
    void reset();

    /*!
     * \return fraction of max distance to the closest object in the ray of the sensor.
     */
    float fraction() const;

    /*!
     * \return the physical model of the closest object hit by the ray of the sensor.
     */
    b2Body *objectHit() const;
private:
    float _fraction;
    b2Body *_object;
};

/*!
 * \brief The ChirpIrSensor class represents the distance sensors on the Chirp robot.
 *
 * These are infrared LEDs with a phototransistor to measure reflected light from the LEDs. It gives a measure of distance from an object, but requires
 * close proximity before the value changes much.
 */
class ChirpIrSensor : public Sensor
{
public:
    explicit ChirpIrSensor(PhysicalObject *robot, QPointF position,
                      float rotationAngle, float reach, float maxValue, float sensorWidth, uint rays);

    /*!
     * \brief Gets the distance value of the sensor.
     * \return The distance value of the sensor.
     */
    unsigned value();

    /*!
     * \brief Gets the closest object hit by the ray.
     * \return The closest object hit by the ray.
     */
    PhysicalObject *objectHit();

    /*!
     * \brief Called by the parent robot each step to update the value of the sensor.
     */
    void update();

private:
    PhysicalObject *_parentBody;

    QPointF _position;
    float _maxDistance;
    float _maxValue;
    QVector<float> _fractions;
    float _baseAngle;

    float _sensorWidth;
    uint _rays;

    IrSensorRayCastCallback _callback;
};


#endif // IRSENSOR_H
