#ifndef WORLD_H
#define WORLD_H

#include <QGraphicsScene>
#include <QList>

class PhysicalObject;
class b2World;
class QJsonDocument;
/*!
 * \brief The World class represents the physical world being simulated.
 *
 * It has methods for adding and removing objects from the simulation.
 * The step function will call the step-method on all PhysicalObject stored in the simulation.
 *
 * The world can be created from a JSON-file, which will delete the old world and create a new one based on the objects of the JSON-file.
 */
class World : public QGraphicsScene
{
public:
    ~World();

    /*!
     * \brief Adds an object to the simulation.
     * \param object PhysicalObject that will be added to the simulation.
     */
    void addObject(PhysicalObject *object);

    /*!
     * \brief Removes an object from the simulation.
     * \param object PhysicalObject that will be removed from the simulation.
     */
    void removeObject(PhysicalObject *object);

    /*!
     * \brief Steps the simulation once, then calls the step method of all the objects in the simulation.
     * \param dt Time in seconds since last step.
     */
    void step(float dt);

    /*!
     * \brief Called by SimulatorView to synchronize the graphical view (positions and rotations) with the
     *        physical simulation of each object.
     */
    void stepGraphics();

    /*!
     * \brief Getter for the Box2D world object.
     * \return The Box2D world object holding all the simulation details.
     */
    b2World *world();

    /*!
     * \brief Named constructor creating a world with many PhysicalObject based on a JSON-document.
     * \param doc The JSON-document containing the objects in the world.
     * \return The newly created world.
     */
    static World *fromJsonDocument(const QJsonDocument &doc);

    /*!
     * \brief Static getter method to get the singleton instance of the class.
     * \return Singleton instance of World.
     */
    static World *instance();
signals:
    
public slots:
    
private:
    explicit World();
    static World *_instance;

    b2World *_world;
    QList<PhysicalObject*> _physicalObjects;
};

#endif // WORLD_H
