#include "irsensor.h"

#include <Box2D/Box2D.h>
#include <QPainter>
#include <QStyleOptionGraphicsItem>
#include <QWidget>
#include <QDebug>
#include <QJsonObject>
#include <cmath>

#include "drawutilities.h"
#include "mathutility.h"
#include "sensor.h"
#include "world.h"

ChirpIrSensor::ChirpIrSensor(PhysicalObject *robot, QPointF position, float rotationAngle,
                   float reach, float maxValue, float sensorWidth, uint rays) : Sensor()
{
    _fractions = QVector<float>(rays, 1);

    _parentBody = robot;
    _position = position;
    _baseAngle = rotationAngle * DEGTORAD;
    _maxDistance = reach;
    _maxValue = maxValue;

    _sensorWidth = sensorWidth * DEGTORAD;
    _rays = rays;
}

unsigned ChirpIrSensor::value()
{
    float minFraction = std::numeric_limits<float>::max();
    foreach (float fraction, _fractions)
    {
        if(fraction < minFraction)
            minFraction = fraction;
    }

    // Clamp the factor between 0.0f and 1.0f
    float factor = clamp(powf(minFraction - 1.0f, 4), 0.0f, 1.0f);
    return _maxValue * factor;

}

PhysicalObject *ChirpIrSensor::objectHit()
{
    return static_cast<PhysicalObject*>(_callback.objectHit()->GetUserData());
}

void ChirpIrSensor::update()
{
    float startAngle = _baseAngle - (_sensorWidth / 2.0f);
    float angleStep = _sensorWidth / _rays;
    for(uint i = 0; i < _rays; ++i)
    {
        float angle = startAngle + (angleStep * i);
        _callback.reset();
        b2Vec2 outerPos = b2Vec2(_position.x(), _position.y()) + _maxDistance * b2Vec2(sinf(angle), cosf(angle));
        _parentBody->body()->GetWorld()->RayCast(&_callback, _parentBody->body()->GetWorldPoint(b2Vec2(_position.x(), _position.y())), _parentBody->body()->GetWorldPoint(outerPos));
        _fractions[i] = _callback.fraction();
    }
}


float32 IrSensorRayCastCallback::ReportFixture(b2Fixture *fixture, const b2Vec2 &point, const b2Vec2 &normal, float32 fraction)
{
    Q_UNUSED(normal);
    Q_UNUSED(point);
    Q_UNUSED(fixture);

    _fraction = fraction;
    _object = fixture->GetBody();
    return fraction;
}

void IrSensorRayCastCallback::reset()
{
    _fraction = 1;
}

float IrSensorRayCastCallback::fraction() const
{
    return _fraction;
}

b2Body *IrSensorRayCastCallback::objectHit() const
{
    return _object;
}
